/* Manually constructed by concatenating and quoting cert.pem and key.pem.
 * The PEM files were generated using:
 *     openssl req -x509 -newkey rsa:2048 -keyout key.pem -out cert.pem -nodes \
 *         -subj "/C=GB/L=Default City/O=uhttpmock/CN=uhttpmock"
 */
static const gchar *uhm_default_tls_certificate =
	"-----BEGIN CERTIFICATE-----\n"
	"MIIDazCCAlOgAwIBAgIJAJKDBYaIm2SBMA0GCSqGSIb3DQEBBQUAMEwxCzAJBgNV\n"
	"BAYTAlVLMRUwEwYDVQQHDAxEZWZhdWx0IENpdHkxEjAQBgNVBAoMCXVodHRwbW9j\n"
	"azESMBAGA1UEAwwJdWh0dHBtb2NrMB4XDTEzMDgwNTA5MTgwOVoXDTEzMDkwNDA5\n"
	"MTgwOVowTDELMAkGA1UEBhMCVUsxFTATBgNVBAcMDERlZmF1bHQgQ2l0eTESMBAG\n"
	"A1UECgwJdWh0dHBtb2NrMRIwEAYDVQQDDAl1aHR0cG1vY2swggEiMA0GCSqGSIb3\n"
	"DQEBAQUAA4IBDwAwggEKAoIBAQDfAjSOnRNZlDGN21zeSsdqt8efsHJVVEKq3iC/\n"
	"hgYQ5lLQKeSF7z4c+uCvWYEOwj6gWxQG8Nzp6pxnHkGoextAkoaOFM6I1OKB8Hkz\n"
	"fYJo8Q3ax9GoykxCFJRlw5ceQ67rYkQOR346xgxyN2Y8fCWO7kVNzQFFFA1dPcTJ\n"
	"Wy9OMHtRPiLmS8CW3QZrFFijyWHrkcQ7GrxYDYMLQ8iso9jdcwv2US0B0Msmxlr7\n"
	"VWzhaZPhr2PaWQMk21j7w1rAuYng7vuWEkLWCTcy+N83qUP+DPWZemso0VuFhanw\n"
	"g1etnhVFCw9wsxIoXLzijRj+/JLuLCWlPAiXtsnVxo/5Y+ytAgMBAAGjUDBOMB0G\n"
	"A1UdDgQWBBRvMD4pnfrMgmzOftZMBtIMGHdcnjAfBgNVHSMEGDAWgBRvMD4pnfrM\n"
	"gmzOftZMBtIMGHdcnjAMBgNVHRMEBTADAQH/MA0GCSqGSIb3DQEBBQUAA4IBAQB+\n"
	"OcT8GNwlxAraicePg/C6N7ZYvxxm4xAWpID0nD5I5jfJw7LuyTPUHoVPBOYE6bW6\n"
	"IyhScGG9TiVCZISMWZGJ5oTjm9n1txRcUOsXHnROFSVN2FWgXIx4ep2MQWMB4WSF\n"
	"ZDzSjCt706vejUyh+HkAWkswNtoKpn3T7vSVg25yJE8k4D120rru9iMaXzvIZad1\n"
	"kn3YQoy3X+tAxlVMTWXn44iswECvrOYf5yxGMi/+AppkEGfWIH8pqUAAvbTqOk9P\n"
	"9L4Bu1AIXfF/rLv8VlU8zojQ70MKOr039TvjuCOCUMX9tXqAC4jYwiIAF/dfRaHe\n"
	"vsDxdSKLB/1rR1vCXLhD\n"
	"-----END CERTIFICATE-----"
	"\n"
	"-----BEGIN PRIVATE KEY-----\n"
	"MIIEvgIBADANBgkqhkiG9w0BAQEFAASCBKgwggSkAgEAAoIBAQDfAjSOnRNZlDGN\n"
	"21zeSsdqt8efsHJVVEKq3iC/hgYQ5lLQKeSF7z4c+uCvWYEOwj6gWxQG8Nzp6pxn\n"
	"HkGoextAkoaOFM6I1OKB8HkzfYJo8Q3ax9GoykxCFJRlw5ceQ67rYkQOR346xgxy\n"
	"N2Y8fCWO7kVNzQFFFA1dPcTJWy9OMHtRPiLmS8CW3QZrFFijyWHrkcQ7GrxYDYML\n"
	"Q8iso9jdcwv2US0B0Msmxlr7VWzhaZPhr2PaWQMk21j7w1rAuYng7vuWEkLWCTcy\n"
	"+N83qUP+DPWZemso0VuFhanwg1etnhVFCw9wsxIoXLzijRj+/JLuLCWlPAiXtsnV\n"
	"xo/5Y+ytAgMBAAECggEBAKyqEnNRHsg+u1dwPqlGELyJ6p+zh/KVOMD/Fy/y3wN6\n"
	"sSfHJ86je94ISKq7i/cXYlHZ8tmk6Aacxdf5TzDZkDIgwNmFNpwu5+lffRfm12/V\n"
	"I7r+90/OwdhwAq4AECM4mFhbTwIXTJ7I9J4CUsAGBzZxOWuRjKglbGuDBbg5R5kq\n"
	"8kPWMGZHT/Osiivaa6f19DKNbQace8DBfqcescGhDaXXMTZnSishEik34uPXI4rm\n"
	"O+qBk/jbPKHIb8HKm23QcM1l4e1m++0wLjeSOBDzq6O8mHgV+GAo57uZmSdnJ2mm\n"
	"OMIiSuVdhDyNUfbIFqgko3tgH37EhwufHlFx0WOytwECgYEA9mWCv2IjvhuTXAL5\n"
	"hR/f5whiO26O3asELwUcYwO1V8/8mFvPmVjwjVpgBNCNr50r7VDZvSJoJ03YisO2\n"
	"PZNNWh9GFafA54NcmHjpzbT4zxYj3dVw9vzPhOV9ndCZWmWhNangMq/xFnUFixjL\n"
	"VpqlqHb/SkdUK6mT3C4FEJmUg00CgYEA57NV0ai+ugx5Wl5lHPsE+aYZ5JRv2OsQ\n"
	"qB+N63msxQvdPFnvoN2fH1g1zUi1MlA6q38EicwyLPNJ36Iy7RXKuBY5B+X6iI6E\n"
	"4ziB/5Bp0mUo5HvPAjjQusiWaslT9AOojA6VqZtuiewkExpl4VkC/0BbWTKtzfnH\n"
	"nzNfxOifnuECgYEA8IJHrM57+x5pqb/Rlet2H8rkMAUL+T3seIUxn0jIY23Gr2W2\n"
	"74WMUT3tSeXU153AegaYc3C5X9wxycmeAt7c+2JZg9vahWGJKd+kwMGsuF9xJSEq\n"
	"Ajzlx2BHTJuIhV98i6hFCtUIrJYPkiXinaeYmieFrBiSBYiipqJZGUoAWGUCgYAF\n"
	"cOWsBb1s2wwifiL7uj1Uq3ziLEYwHt0GRa9sfy/6dJveZfJFoc6xyr2As5tlshKe\n"
	"ol316nCnM5NhiAqQHLnk9siiEdl/SXF/cH1FBhwmD7AVJX8n+zOTn1BA87df/JIB\n"
	"r/n9wKOo4325YR5RW2jBm75JavI/6wSwDWHLWvccoQKBgEwjhN5yW8tqHl6BUtn+\n"
	"brqvwk3azURyy5A9GlgYifVprQ2qAnxJZxYibcBmTOH3ewDfmWB3t7piHy028E9y\n"
	"sXEnj6LaYbfLX36Pp8GgGKDGHxwjviMGWEW8nnrZgAvtWm7DRSjzLzeajYL3SE5a\n"
	"3IrgLWcOTfnUg6FzQ6xAAlxa\n"
	"-----END PRIVATE KEY-----"
;
